
from pathlib import Path
from typing import Union


class Provider:

    def __init__(self, project_root_dir: Union[str, Path]):
        """Creates the provider instance

        Args:
            project_root_dir: the root directory of the project
        """
        self.project_root_dir = Path(project_root_dir)

    def build(self, *args, **kwargs):
        raise NotImplementedError("Child classes must implement the build method.")

    def run(self, *args, **kwargs):
        raise NotImplementedError("Child classes must implement the run method.")


class CloudMixin:

    def deploy(self, *args, **kwargs):
        raise NotImplementedError("Child classes must implement the deploy method.")




import os
import yaml

from pathlib import Path
from typing import Union

from bisi.resources import Dockerfile


def create_python_resources(path: Union[str, Path]) -> None:
    """Executes the python file to create the resources defined in it
    The resources will be accessible via Resource.get_resources()

    Args:
        path:
    """
    # TODO: Should we blindly exec the Python file? Is there another way to collect the resource(s)?
    path = Path(path)

    with open(path, 'r') as fp:
        exec(fp.read())


def create_resources(project_root_dir: Union[str, Path]) -> None:
    """Creates the resources defined at a root directory expecting either a yaml configuration or python definition
    The resources will be accessible via Resource.get_resources()

    Args:
        project_root_dir: The path to the root directory to get the resource definitions from
    """
    # TODO: Figure out how to collate these two sources if both exist.
    project_root_dir = Path(project_root_dir)

    python_path = os.path.join(project_root_dir, 'bisi_resources.py')
    if os.path.exists(python_path):
        create_python_resources(python_path)




from typing import Dict, List


class BatchJobConfig:

    def __init__(self, jobQueue: str, jobDefinitionName: str = None, containerProperties: dict = None,
                 extra_args: List[str] = None, environment: Dict[str, str] = None, **kwargs):
        """Config class for configuring batch jobs
        most arguments are passed to https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/batch.html#Batch.Client.register_job_definition

        Args:
            jobQueue: The batch job queue to submit the job to
            jobDefinitionName: the name to give the job definition
            containerProperties: The container properties to set for the job, defaults to 2 vcpus and 1GB of memory,
                when running with bisi, image, command, and environment are overridden
            extra_args: a list of extra arguments to call the entrypoint with
            environment: a dictionary of environment variables to set for batch jobs
            **kwargs: extra args to pass to boto3.client('batch').register_job_definition
        """
        self.jobQueue = jobQueue
        self.jobDefinitionName = jobDefinitionName
        self.type = 'container'
        if containerProperties is None:
            self.containerProperties = {
                'vcpus': 2,
                'memory': 1024
            }
        else:
            self.containerProperties = containerProperties

        self.extra_args = extra_args if extra_args else list()
        self.environment = environment if environment else dict()
        self.kwargs = kwargs




from typing import Union, Dict

from bisi.resources import Resource, Dockerfile
from bisi.resources.config import LocalJobConfig, BatchJobConfig, ECRConfig
from bisi.utils import get_git_branch, clean_string, get_project_root_dir


class Job(Resource):

    def __init__(self, name: str, entrypoint: str, dockerfile: Union[str, Dockerfile] = None,
                 environment: Dict[str, str] = None, local_config: Union[dict, LocalJobConfig] = None,
                 batch_config: Union[dict, BatchJobConfig] = None):
        """Creates a resource representing a job

        Args:
            name: the name of this dockerfile
            entrypoint: relative path to the dockerfile from the root of the project
            dockerfile: the name of the dockerfile to run in, required if there are multiple dockerfiles defined
            environment: a dictionary of environment variables to run the job with
            local_config: a LocalJobConfig object to use when running this job locally
            batch_config: a BatchJobConfig object to use when running this job in AWS Batch
        """
        super().__init__()
        self.name = name
        self.entrypoint = entrypoint

        if dockerfile is None:
            dockerfiles = self.get_resources(resource_type=Dockerfile)
        elif isinstance(dockerfile, str):
            dockerfiles = self.get_resources(resource_type=Dockerfile, name=dockerfile)
        else:
            dockerfiles = [dockerfile]

        if len(dockerfiles) == 0:
            raise Exception('Must specify at least one dockerfile resource in your resources')
        if len(dockerfiles) > 1:
            raise Exception('Must specify the dockerfile name if more than one dockerfile is defined')

        self.dockerfile = dockerfiles[0]

        self.environment = environment if environment is not None else dict()
        self.local_config = LocalJobConfig(**local_config) if isinstance(local_config, dict) else local_config
        self.batch_config = BatchJobConfig(**batch_config) if isinstance(batch_config, dict) else batch_config

        if self.batch_config is not None:
            if self.batch_config.jobDefinitionName is None:
                self.batch_config.jobDefinitionName = f'{self.dockerfile.name}-{self.name}'
                branch = get_git_branch()
                if branch is not None:
                    self.batch_config.jobDefinitionName += f'-{clean_string(branch)}'
            if self.dockerfile.ecr_config is None:
                self.dockerfile.ecr_config = ECRConfig(repositoryName=get_project_root_dir().name)

        if self.local_config is None:
            self.local_config = LocalJobConfig()

    def __repr__(self):
        return f'Job(name={self.name}, entrypoint={self.entrypoint}, dockerfile={self.dockerfile})'




import os
import sys
import docker

from typing import Iterable

from bisi.providers import Provider
from bisi.resources import Resource, Dockerfile, Job

docker_client = docker.from_env()


class Local(Provider):

    def build(self) -> None:
        """Builds the resources that are specified locally"""
        dockerfiles = Resource.get_resources(resource_type=Dockerfile)
        for dockerfile in dockerfiles:
            self.build_dockerfile(dockerfile)

    def deploy(self) -> None:
        pass  # stub to make project work

    def build_dockerfile(self, dockerfile: Dockerfile) -> None:
        """Builds the dockerfile locally using docker

        Args:
            dockerfile: the dockerfile to build
        """
        print(f'Building {dockerfile}')
        build_args = ' '.join([f'--build-arg "{k}={dockerfile.build_args[k]}"' for k in dockerfile.build_args]) \
            if dockerfile.build_args is not None else ''
        secrets = ' '.join([f'--secret "{secret}"' for secret in dockerfile.secret]) \
            if dockerfile.secret is not None else ''
        ssh = f'--ssh {dockerfile.ssh}' if dockerfile.ssh is not None else ''

        ret_val = os.system(
            f'docker build '
            f'-f {os.path.join(self.project_root_dir, dockerfile.file)} '
            f'-t {dockerfile.latest_image_tag()} '
            f'{build_args} '
            f'{secrets} '
            f'{ssh} '
            f'{self.project_root_dir} '
        )

        if ret_val != 0:
            raise Exception(f'Failed to build Dockerfile, exit_code = {ret_val}')

        for tag in dockerfile.tags[1:]:
            ret_val = os.system(f'docker tag {dockerfile.name}:{dockerfile.tags[0]} {dockerfile.name}:{tag}')
            if ret_val != 0:
                raise Exception(f'Failed to tag docker image')

    def run(self, name: str, arguments: Iterable[str]):
        """Runs the given resource with the provided arguments

        Args:
            name: the name of the resource to run
            arguments: the arguments to run with
        """
        job = Resource.get_resources(resource_type=Job, name=name)
        if len(job) != 1:
            raise Exception(f'Could not find job with name={name}')
        self.run_job(job[0], arguments)

    @staticmethod
    def run_job(job: Job, arguments: Iterable[str]):
        """Runs the job with the specified arguments locally with py-docker

        Args:
            job: the job to run
            arguments: the arguments to run with
        """
        print(f'Running {job}')

        command = [job.entrypoint] + list(arguments) + job.local_config.extra_args
        if job.entrypoint.endswith('.py'):
            command = ['python'] + command

        environment = job.environment.copy()
        environment.update(job.local_config.environment)

        try:
            container = docker_client.containers.run(
                image=job.dockerfile.latest_image_tag(),
                command=command,
                environment=environment,
                remove=job.local_config.remove,
                stream=True,
                **job.local_config.kwargs
            )
            for log in container:
                print(log.decode('utf-8').strip('\n'))
        except Exception as e:
            print('Job Failed', file=sys.stderr)
            raise e
